import logging
import os
from pathlib import Path
from types import SimpleNamespace

import numpy as np
import volume_segmantics.utilities.base_data_utils as utils
from skimage import img_as_ubyte, io
from tqdm import tqdm
from volume_segmantics.data.base_data_manager import BaseDataManager
from typing import Union


class TrainingDataSlicer(BaseDataManager):
    """
    Class that performs image preprocessing and provides methods to 
    convert 3d data volumes into 2d image slices on disk for model training.
    Slicing is carried in all of the xy (z), xz (y) and yz (x) planes.
    """

    def __init__(
        self,
        data_vol: Union[str, np.ndarray],
        label_vol: Union[str, np.ndarray],
        settings: SimpleNamespace,
    ):
        """Inits TrainingDataSlicer.

        Args:
            data_vol(Union[str, np.ndarray]): Either a path to an image data volume or a numpy array of 3D image data
            label_vol(Union[str, np.ndarray]): Either a path to a label data volume or a numpy array of 3D label data
            settings(SimpleNamespace): An object containing the training settings
        """
        super().__init__(data_vol, settings)
        self.data_im_out_dir = None
        self.seg_im_out_dir = None
        self.multilabel = False
        self.settings = settings
        self.label_vol_path = utils.setup_path_if_exists(label_vol)
        if self.label_vol_path is not None:
            self.seg_vol, _ = utils.get_numpy_from_path(
                self.label_vol_path, internal_path=settings.seg_hdf5_path
            )
        elif isinstance(label_vol, np.ndarray):
            self.seg_vol = label_vol
        self._preprocess_labels()

    def _preprocess_labels(self):
        seg_classes = np.unique(self.seg_vol)
        self.num_seg_classes = len(seg_classes)
        if self.num_seg_classes > 2:
            self.multilabel = True
        logging.info(
            "Number of classes in segmentation dataset:" f" {self.num_seg_classes}"
        )
        logging.info(f"These classes are: {seg_classes}")
        if seg_classes[0] != 0 or not utils.sequential_labels(seg_classes):
            logging.info("Fixing label classes.")
            self._fix_label_classes(seg_classes)
        self.codes = [f"label_val_{i}" for i in seg_classes]

    def _fix_label_classes(self, seg_classes):
        """Changes the data values of classes in a segmented volume so that
        they start from zero.

        Args:
            seg_classes(list): An ascending list of the labels in the volume.
        """
        for idx, current in enumerate(seg_classes):
            self.seg_vol[self.seg_vol == current] = idx

    def output_data_slices(self, data_dir: Path, prefix: str) -> None:
        """
        Method that triggers slicing image data volume to disk in the
        xy (z), xz (y) and yz (x) planes.

        Args:
            data_dir (Path): Path to the directory for image output
            prefix (str): String to prepend to image filename
        """
        self.data_im_out_dir = data_dir
        logging.info("Slicing data volume and saving slices to disk")
        os.makedirs(data_dir, exist_ok=True)
        self._output_slices_to_disk(self.data_vol, data_dir, prefix)

    def output_label_slices(self, data_dir: Path, prefix: str) -> None:
        """
        Method that triggers slicing label data volume to disk in the
        xy (z), xz (y) and yz (x) planes.

        Args:
            data_dir (Path): Path to the directory for label image output
            prefix (str): String to prepend to image filename
        """
        self.seg_im_out_dir = data_dir
        logging.info("Slicing label volume and saving slices to disk")
        os.makedirs(data_dir, exist_ok=True)
        self._output_slices_to_disk(self.seg_vol, data_dir, prefix, label=True)

    def _output_slices_to_disk(self, data_arr, output_path, name_prefix, label=False):
        """Coordinates the slicing of an image volume in the three orthogonal
        planes to images on disk.

        Args:
            data_arr (array): The data volume to be sliced.
            output_path (pathlib.Path): A Path object to the output directory.
            label (bool): Whether this is a label volume.
        """
        shape_tup = data_arr.shape
        axis_enum = utils.get_training_axis(self.settings)
        ax_idx_pairs = utils.get_axis_index_pairs(shape_tup, axis_enum)
        num_ims = utils.get_num_of_ims(shape_tup, axis_enum)
        for axis, index in tqdm(ax_idx_pairs, total=num_ims):
            out_path = output_path / f"{name_prefix}_{axis}_stack_{index}"
            self._output_im(
                utils.axis_index_to_slice(data_arr, axis, index), out_path, label
            )

    def _output_im(self, data, path, label=False):
        """Converts a slice of data into an image on disk.

        Args:
            data (numpy.array): The data slice to be converted.
            path (str): The path of the image file including the filename prefix.
            label (bool): Whether to convert values >1 to 1 for binary segmentation.
        """
        # TODO: Allow saving a higher bit depth
        if data.dtype != np.uint8:
            data = img_as_ubyte(data)

        if label and not self.multilabel:
            data[data > 1] = 1
        io.imsave(f"{path}.png", data, check_contrast=False)

    def _delete_image_dir(self, im_dir_path):
        if im_dir_path.exists():
            ims = list(im_dir_path.glob("*.png"))
            logging.info(f"Deleting {len(ims)} images.")
            for im in ims:
                im.unlink()
            logging.info(f"Deleting the empty directory.")
            im_dir_path.rmdir()

    def clean_up_slices(self) -> None:
        """
        Deletes data and label image slices created by Slicer.
        """
        self._delete_image_dir(self.data_im_out_dir)
        self._delete_image_dir(self.seg_im_out_dir)



import csv
import logging
import math
from pathlib import Path
import sys
import time
from types import SimpleNamespace
from typing import Union

import matplotlib as mpl
import numpy as np
import termplotlib as tpl

mpl.use("Agg")
import torch
import torch.nn as nn
import volume_segmantics.utilities.base_data_utils as utils
import volume_segmantics.utilities.config as cfg
from matplotlib import pyplot as plt
from tqdm import tqdm
from volume_segmantics.data.dataloaders import get_2d_training_dataloaders
from volume_segmantics.data.pytorch3dunet_losses import (
    BCEDiceLoss,
    DiceLoss,
    GeneralizedDiceLoss,
)
from volume_segmantics.data.pytorch3dunet_metrics import (
    DiceCoefficient,
    MeanIoU,
)
from volume_segmantics.model.model_2d import create_model_on_device
from volume_segmantics.utilities.early_stopping import EarlyStopping


class VolSeg2dTrainer:
    """Class that provides methods to train a 2d deep learning model
    and to output figures showing training and validation loss and
    example predictions on random validation images.
    """

    def __init__(
        self,
        image_dir_path: Path,
        label_dir_path: Path,
        labels: Union[int, dict],
        settings: SimpleNamespace,
    ):
        """Inits VolSeg2dTrainer.

        Args:
            image_dir_path (Path): Path to directory containing image data slices.
            label_dir_path (Path): Path to directory containing label data slices.
            labels (Union[int, dict]): Either number of labels or dictionary containing label names and associated metadata.
            settings (SimpleNamespace): A training settings object.
        """
        self.training_loader, self.validation_loader = get_2d_training_dataloaders(
            image_dir_path, label_dir_path, settings
        )
        self.label_no = labels if isinstance(labels, int) else len(labels)
        self.codes = labels if isinstance(labels, dict) else {}
        self.settings = settings
        # Params for learning rate finder
        self.starting_lr = float(settings.starting_lr)
        self.end_lr = float(settings.end_lr)
        self.log_lr_ratio = self._calculate_log_lr_ratio()
        self.lr_find_epochs = settings.lr_find_epochs
        self.lr_reduce_factor = settings.lr_reduce_factor
        # Params for model training
        self.model_device_num = int(settings.cuda_device)
        self.patience = settings.patience
        self.loss_criterion = self._get_loss_criterion()
        self.eval_metric = self._get_eval_metric()
        self.model_struc_dict = self._get_model_struc_dict(settings)
        self.avg_train_losses = []  # per epoch training loss
        self.avg_valid_losses = []  #  per epoch validation loss
        self.avg_eval_scores = []  #  per epoch evaluation score

    def _get_model_struc_dict(self, settings):
        model_struc_dict = settings.model
        model_type = utils.get_model_type(settings)
        model_struc_dict["type"] = model_type
        model_struc_dict["in_channels"] = cfg.MODEL_INPUT_CHANNELS
        model_struc_dict["classes"] = self.label_no
        return model_struc_dict

    def _calculate_log_lr_ratio(self):
        return math.log(self.end_lr / self.starting_lr)

    def _create_model_and_optimiser(self, learning_rate, frozen=False):
        logging.info(f"Setting up the model on device {self.settings.cuda_device}.")
        self.model = create_model_on_device(
            self.model_device_num, self.model_struc_dict
        )
        if frozen:
            self._freeze_model()
        logging.info(
            f"Model has {self._count_trainable_parameters()} trainable parameters, {self._count_parameters()} total parameters."
        )
        self.optimizer = self._create_optimizer(learning_rate)
        logging.info("Trainer created.")

    def _freeze_model(self):
        logging.info(
            f"Freezing model with {self._count_trainable_parameters()} trainable parameters, {self._count_parameters()} total parameters."
        )
        for name, param in self.model.named_parameters():
            if all(["encoder" in name, "conv" in name]) and param.requires_grad:
                param.requires_grad = False

    def _unfreeze_model(self):
        logging.info(
            f"Unfreezing model with {self._count_trainable_parameters()} trainable parameters, {self._count_parameters()} total parameters."
        )
        for name, param in self.model.named_parameters():
            if all(["encoder" in name, "conv" in name]) and not param.requires_grad:
                param.requires_grad = True

    def _count_trainable_parameters(self) -> int:
        return sum(p.numel() for p in self.model.parameters() if p.requires_grad)

    def _count_parameters(self) -> int:
        return sum(p.numel() for p in self.model.parameters())

    def _get_loss_criterion(self):
        if self.settings.loss_criterion == "BCEDiceLoss":
            alpha = self.settings.alpha
            beta = self.settings.beta
            logging.info(
                f"Using combined BCE and Dice loss with weighting of {alpha}*BCE "
                f"and {beta}*Dice"
            )
            loss_criterion = BCEDiceLoss(alpha, beta)
        elif self.settings.loss_criterion == "DiceLoss":
            logging.info("Using DiceLoss")
            loss_criterion = DiceLoss(normalization="none")
        elif self.settings.loss_criterion == "BCELoss":
            logging.info("Using BCELoss")
            loss_criterion = nn.BCEWithLogitsLoss()
        elif self.settings.loss_criterion == "CrossEntropyLoss":
            logging.info("Using CrossEntropyLoss")
            loss_criterion = nn.CrossEntropyLoss()
        elif self.settings.loss_criterion == "GeneralizedDiceLoss":
            logging.info("Using GeneralizedDiceLoss")
            loss_criterion = GeneralizedDiceLoss()
        else:
            logging.error("No loss criterion specified, exiting")
            sys.exit(1)
        return loss_criterion

    def _get_eval_metric(self):
        # Get evaluation metric
        if self.settings.eval_metric == "MeanIoU":
            logging.info("Using MeanIoU")
            eval_metric = MeanIoU()
        elif self.settings.eval_metric == "DiceCoefficient":
            logging.info("Using DiceCoefficient")
            eval_metric = DiceCoefficient()
        else:
            logging.error("No evaluation metric specified, exiting")
            sys.exit(1)
        return eval_metric

    def train_model(
        self,
        output_path: Path,
        num_epochs: int,
        patience: int,
        create: bool = True,
        frozen: bool = False,
    ) -> None:
        """Performs training of model for a number of epochs
        with a learning rate that is determined automatically.

        Args:
            output_path (Path): Path to save model file to.
            num_epochs (int): Number of epochs to train the model for.
            patience (int): Number of epochs to wait for while validation loss is not improving before terminating.
            create (bool, optional): Whether to create a new model and optimizer from scratch. Defaults to True.
            frozen (bool, optional): Whether to freeze the weights for convolutional layers in the encoder. Defaults to False.
        """
        train_losses = []
        valid_losses = []
        eval_scores = []

        if create:
            self._create_model_and_optimiser(self.starting_lr, frozen=frozen)
            lr_to_use = self._run_lr_finder()
            # Recreate model and start training
            self._create_model_and_optimiser(lr_to_use, frozen=frozen)
            early_stopping = self._create_early_stopping(output_path, patience)
        else:
            # Reduce starting LR, since model alreadiy partiallly trained
            self.starting_lr /= self.lr_reduce_factor
            self.end_lr /= self.lr_reduce_factor
            self.log_lr_ratio = self._calculate_log_lr_ratio()
            self._load_in_model_and_optimizer(
                self.starting_lr, output_path, frozen=frozen, optimizer=False
            )
            lr_to_use = self._run_lr_finder()
            min_loss = self._load_in_model_and_optimizer(
                self.starting_lr, output_path, frozen=frozen, optimizer=False
            )
            early_stopping = self._create_early_stopping(
                output_path, patience, best_score=-min_loss
            )

        # Initialise the One Cycle learning rate scheduler
        lr_scheduler = self._create_oc_lr_scheduler(num_epochs, lr_to_use)

        for epoch in range(1, num_epochs + 1):
            self.model.train()
            tic = time.perf_counter()
            logging.info(f"Epoch {epoch} of {num_epochs}")
            for batch in tqdm(
                self.training_loader,
                desc="Training batch",
                bar_format=cfg.TQDM_BAR_FORMAT,
            ):
                loss = self._train_one_batch(lr_scheduler, batch)
                train_losses.append(loss.item())  # record training loss

            self.model.eval()  # prep model for evaluation
            with torch.no_grad():
                for batch in tqdm(
                    self.validation_loader,
                    desc="Validation batch",
                    bar_format=cfg.TQDM_BAR_FORMAT,
                ):
                    inputs, targets = utils.prepare_training_batch(
                        batch, self.model_device_num, self.label_no
                    )
                    output = self.model(inputs)  # Forward pass
                    # calculate the loss
                    if self.settings.loss_criterion == "CrossEntropyLoss":
                        loss = self.loss_criterion(output, torch.argmax(targets, dim=1))
                    else:
                        loss = self.loss_criterion(output, targets.float())
                    valid_losses.append(loss.item())  # record validation loss
                    s_max = nn.Softmax(dim=1)
                    probs = s_max(output)  # Convert the logits to probs
                    probs = torch.unsqueeze(probs, 2)
                    targets = torch.unsqueeze(targets, 2)
                    eval_score = self.eval_metric(probs, targets)
                    if eval_score.is_cuda:
                        eval_score = eval_score.cpu().detach().numpy()
                    eval_scores.append(eval_score)  # record eval metric

            toc = time.perf_counter()
            # calculate average loss/metric over an epoch
            self.avg_train_losses.append(np.average(train_losses))
            self.avg_valid_losses.append(np.average(valid_losses))
            self.avg_eval_scores.append(np.average(eval_scores))
            logging.info(
                f"Epoch {epoch}. Training loss: {self.avg_train_losses[-1]}, Validation Loss: "
                f"{self.avg_valid_losses[-1]}. {self.settings.eval_metric}: {self.avg_eval_scores[-1]}"
            )
            logging.info(f"Time taken for epoch {epoch}: {toc - tic:0.2f} seconds")
            # clear lists to track next epoch
            train_losses = []
            valid_losses = []
            eval_scores = []

            # early_stopping needs the validation loss to check if it has decreased,
            # and if it has, it will make a checkpoint of the current model
            early_stopping(
                self.avg_valid_losses[-1], self.model, self.optimizer, self.codes
            )

            if early_stopping.early_stop:
                logging.info("Early stopping")
                break

        # load the last checkpoint with the best model
        self._load_in_weights(output_path)

    def _load_in_model_and_optimizer(
        self, learning_rate, output_path, frozen=False, optimizer=False
    ):
        self._create_model_and_optimiser(learning_rate, frozen=frozen)
        logging.info("Loading in weights from saved checkpoint.")
        loss_val = self._load_in_weights(output_path, optimizer=optimizer)
        return loss_val

    def _load_in_weights(self, output_path, optimizer=False, gpu=True):
        # load the last checkpoint with the best model
        if gpu:
            map_location = f"cuda:{self.model_device_num}"
        else:
            map_location = "cpu"
        model_dict = torch.load(output_path, map_location=map_location)
        logging.info("Loading model weights.")
        self.model.load_state_dict(model_dict["model_state_dict"])
        if optimizer:
            logging.info("Loading optimizer weights.")
            self.optimizer.load_state_dict(model_dict["optimizer_state_dict"])
        return model_dict.get("loss_val", np.inf)

    def _run_lr_finder(self):
        logging.info("Finding learning rate for model.")
        lr_scheduler = self._create_exponential_lr_scheduler()
        lr_find_loss, lr_find_lr = self._lr_finder(lr_scheduler)
        lr_to_use = self._find_lr_from_graph(lr_find_loss, lr_find_lr)
        logging.info(f"LR to use {lr_to_use}")
        return lr_to_use

    def _lr_finder(self, lr_scheduler, smoothing=0.05):
        lr_find_loss = []
        lr_find_lr = []
        iters = 0

        self.model.train()
        logging.info(
            f"Training for {self.lr_find_epochs} epochs to create a learning "
            "rate plot."
        )
        for i in range(self.lr_find_epochs):
            for batch in tqdm(
                self.training_loader,
                desc=f"Epoch {i + 1}, batch number",
                bar_format=cfg.TQDM_BAR_FORMAT,
            ):
                loss = self._train_one_batch(lr_scheduler, batch)
                lr_step = self.optimizer.state_dict()["param_groups"][0]["lr"]
                lr_find_lr.append(lr_step)
                if iters == 0:
                    lr_find_loss.append(loss)
                else:
                    loss = smoothing * loss + (1 - smoothing) * lr_find_loss[-1]
                    lr_find_loss.append(loss)
                if loss > 1 and iters > len(self.training_loader) // 1.333:
                    break
                iters += 1

        if self.settings.plot_lr_graph:
            fig = tpl.figure()
            fig.plot(
                np.log10(lr_find_lr),
                lr_find_loss,
                width=50,
                height=30,
                xlabel="Log10 Learning Rate",
            )
            fig.show()

        return lr_find_loss, lr_find_lr

    @staticmethod
    def _find_lr_from_graph(
        lr_find_loss: torch.Tensor, lr_find_lr: torch.Tensor
    ) -> float:
        """Calculates learning rate corresponsing to minimum gradient in graph
        of loss vs learning rate.

        Args:
            lr_find_loss (torch.Tensor): Loss values accumulated during training
            lr_find_lr (torch.Tensor): Learning rate used for mini-batch

        Returns:
            float: The learning rate at the point when loss was falling most steeply
            divided by a fudge factor.
        """
        default_min_lr = cfg.DEFAULT_MIN_LR  # Add as default value to fix bug
        # Get loss values and their corresponding gradients, and get lr values
        for i in range(0, len(lr_find_loss)):
            if lr_find_loss[i].is_cuda:
                lr_find_loss[i] = lr_find_loss[i].cpu()
            lr_find_loss[i] = lr_find_loss[i].detach().numpy()
        losses = np.array(lr_find_loss)
        try:
            gradients = np.gradient(losses)
            min_gradient = gradients.min()
            if min_gradient < 0:
                min_loss_grad_idx = gradients.argmin()
            else:
                logging.info(
                    f"Minimum gradient: {min_gradient} was positive, returning default value instead."
                )
                return default_min_lr
        except Exception as e:
            logging.info(f"Failed to compute gradients, returning default value. {e}")
            return default_min_lr
        min_lr = lr_find_lr[min_loss_grad_idx]
        return min_lr / cfg.LR_DIVISOR

    def _lr_exp_stepper(self, x):
        """Exponentially increase learning rate as part of strategy to find the
        optimum.
        Taken from
        https://towardsdatascience.com/adaptive-and-cyclical-learning-rates-using-pytorch-2bf904d18dee
        """
        return math.exp(
            x * self.log_lr_ratio / (self.lr_find_epochs * len(self.training_loader))
        )

    def _create_optimizer(self, learning_rate):
        return torch.optim.AdamW(self.model.parameters(), lr=learning_rate)

    def _create_exponential_lr_scheduler(self):
        return torch.optim.lr_scheduler.LambdaLR(self.optimizer, self._lr_exp_stepper)

    def _create_oc_lr_scheduler(self, num_epochs, lr_to_use):
        return torch.optim.lr_scheduler.OneCycleLR(
            self.optimizer,
            max_lr=lr_to_use,
            steps_per_epoch=len(self.training_loader),
            epochs=num_epochs,
            pct_start=self.settings.pct_lr_inc,
        )

    def _create_early_stopping(self, output_path, patience, best_score=None):
        return EarlyStopping(
            patience=patience,
            verbose=True,
            path=output_path,
            model_dict=self.model_struc_dict,
            best_score=best_score,
        )

    def _train_one_batch(self, lr_scheduler, batch):
        inputs, targets = utils.prepare_training_batch(
            batch, self.model_device_num, self.label_no
        )
        self.optimizer.zero_grad()
        output = self.model(inputs)  # Forward pass
        if self.settings.loss_criterion == "CrossEntropyLoss":
            loss = self.loss_criterion(output, torch.argmax(targets, dim=1))
        else:
            loss = self.loss_criterion(output, targets.float())
        loss.backward()  # Backward pass
        self.optimizer.step()
        lr_scheduler.step()  # update the learning rate
        return loss

    def output_loss_fig(self, model_out_path: Path) -> None:
        """Save out a figure showing training and validation loss versus
        epoch number.

        Args:
            model_out_path (Path): Path to the model output by the trainer. 
        """

        fig = plt.figure(figsize=(10, 8))
        plt.plot(
            range(1, len(self.avg_train_losses) + 1),
            self.avg_train_losses,
            label="Training Loss",
        )
        plt.plot(
            range(1, len(self.avg_valid_losses) + 1),
            self.avg_valid_losses,
            label="Validation Loss",
        )

        minposs = (
            self.avg_valid_losses.index(min(self.avg_valid_losses)) + 1
        )  # find position of lowest validation loss
        plt.axvline(
            minposs, linestyle="--", color="r", label="Early Stopping Checkpoint"
        )

        plt.xlabel("epochs")
        plt.ylabel("loss")
        plt.xlim(0, len(self.avg_train_losses) + 1)  # consistent scale
        plt.grid(True)
        plt.legend()
        plt.tight_layout()
        output_dir = model_out_path.parent
        fig_out_pth = output_dir / f"{model_out_path.stem}_loss_plot.png"
        logging.info(f"Saving figure of training/validation losses to {fig_out_pth}")
        fig.savefig(fig_out_pth, bbox_inches="tight")
        # Output a list of training stats
        epoch_lst = range(len(self.avg_train_losses))
        rows = zip(
            epoch_lst,
            self.avg_train_losses,
            self.avg_valid_losses,
            self.avg_eval_scores,
        )
        with open(output_dir / f"{model_out_path.stem}_train_stats.csv", "w") as f:
            writer = csv.writer(f)
            writer.writerow(("Epoch", "Train Loss", "Valid Loss", "Eval Score"))
            for row in rows:
                writer.writerow(row)

    def output_prediction_figure(self, model_path: Path) -> None:
        """Saves a figure containing image slice data for four random images
        from the validation dataset along with the corresponding ground truth
        label image and corresponding prediction output from the model attached
        to this class instance. The image is saved to the same directory as the
        model weights.

        Args:
            model_path (pathlib.Path): Full path to the model weights file,
            this is used to get the directory and name of the model not to
            load and predict.
        """
        self.model.eval()  # prep model for evaluation
        batch = next(iter(self.validation_loader))  # Get first batch
        with torch.no_grad():
            inputs, targets = utils.prepare_training_batch(
                batch, self.model_device_num, self.label_no
            )
            output = self.model(inputs)  # Forward pass
            s_max = nn.Softmax(dim=1)
            probs = s_max(output)  # Convert the logits to probs
            labels = torch.argmax(probs, dim=1)  # flatten channels

        # Create the plot
        bs = self.validation_loader.batch_size
        if bs < 4:
            rows = bs
        else:
            rows = 4
        fig = plt.figure(figsize=(12, 16))
        columns = 3
        j = 0
        for i in range(columns * rows)[::3]:
            img = inputs[j].squeeze().cpu()
            gt = torch.argmax(targets[j], dim=0).cpu()
            pred = labels[j].cpu()
            col1 = fig.add_subplot(rows, columns, i + 1)
            plt.imshow(img, cmap="gray")
            col2 = fig.add_subplot(rows, columns, i + 2)
            plt.imshow(gt, cmap="gray")
            col3 = fig.add_subplot(rows, columns, i + 3)
            plt.imshow(pred, cmap="gray")
            j += 1
            if i == 0:
                col1.title.set_text("Data")
                col2.title.set_text("Ground Truth")
                col3.title.set_text("Prediction")
        plt.suptitle(f"Predictions for {model_path.name}", fontsize=16)
        plt_out_pth = model_path.parent / f"{model_path.stem}_prediction_image.png"
        logging.info(f"Saving example image predictions to {plt_out_pth}")
        plt.savefig(plt_out_pth, dpi=300)



import logging
from pathlib import Path
from types import SimpleNamespace

import numpy as np
import torch
import volume_segmantics.utilities.base_data_utils as utils
import volume_segmantics.utilities.config as cfg
from torch import nn as nn
from tqdm import tqdm
from volume_segmantics.data.dataloaders import get_2d_prediction_dataloader
from volume_segmantics.model.model_2d import create_model_from_file
from volume_segmantics.utilities.base_data_utils import Axis


class VolSeg2dPredictor:
    """Class that performs U-Net prediction operations. Does not interact with disk."""

    def __init__(self, model_file_path: str, settings: SimpleNamespace) -> None:
        self.model_file_path = Path(model_file_path)
        self.settings = settings
        self.model_device_num = int(settings.cuda_device)
        model_tuple = create_model_from_file(
            self.model_file_path, self.model_device_num
        )
        self.model, self.num_labels, self.label_codes = model_tuple

    def _get_model_from_trainer(self, trainer):
        self.model = trainer.model

    def _predict_single_axis(self, data_vol, output_probs=True, axis=Axis.Z):
        output_vol_list = []
        output_prob_list = []
        data_vol = utils.rotate_array_to_axis(data_vol, axis)
        yx_dims = list(data_vol.shape[1:])
        s_max = nn.Softmax(dim=1)
        data_loader = get_2d_prediction_dataloader(data_vol, self.settings)
        self.model.eval()
        logging.info(f"Predicting segmentation for volume of shape {data_vol.shape}.")
        with torch.no_grad():
            for batch in tqdm(
                data_loader, desc="Prediction batch", bar_format=cfg.TQDM_BAR_FORMAT
            ):
                output = self.model(batch.to(self.model_device_num))  # Forward pass
                probs = s_max(output)  # Convert the logits to probs
                # TODO: Don't flatten channels if one-hot output is needed
                labels = torch.argmax(probs, dim=1)  # flatten channels
                labels = utils.crop_tensor_to_array(labels, yx_dims)
                output_vol_list.append(labels.astype(np.uint8))
                if output_probs:
                    # Get indices of max probs
                    max_prob_idx = torch.argmax(probs, dim=1, keepdim=True)
                    # Extract along axis from outputs
                    probs = torch.gather(probs, 1, max_prob_idx)
                    # Remove the label dimension
                    probs = torch.squeeze(probs, dim=1)
                    probs = utils.crop_tensor_to_array(probs, yx_dims)
                    output_prob_list.append(probs.astype(np.float16))

        labels = np.concatenate(output_vol_list)
        labels = utils.rotate_array_to_axis(labels, axis)
        probs = np.concatenate(output_prob_list) if output_prob_list else None
        if probs is not None:
            probs = utils.rotate_array_to_axis(probs, axis)
        return labels, probs

    def _predict_3_ways_max_probs(self, data_vol):
        shape_tup = data_vol.shape
        logging.info("Creating empty data volumes in RAM to combine 3 axis prediction.")
        label_container = np.empty((2, *shape_tup), dtype=np.uint8)
        prob_container = np.empty((2, *shape_tup), dtype=np.float16)
        logging.info("Predicting YX slices:")
        label_container[0], prob_container[0] = self._predict_single_axis(
            data_vol, output_probs=True
        )
        logging.info("Predicting ZX slices:")
        label_container[1], prob_container[1] = self._predict_single_axis(
            data_vol, output_probs=True, axis=Axis.Y
        )
        logging.info("Merging XY and ZX volumes.")
        self._merge_vols_in_mem(prob_container, label_container)
        logging.info("Predicting ZY slices:")
        label_container[1], prob_container[1] = self._predict_single_axis(
            data_vol, output_probs=True, axis=Axis.X
        )
        logging.info("Merging max of XY and ZX volumes with ZY volume.")
        self._merge_vols_in_mem(prob_container, label_container)
        return label_container[0], prob_container[0]

    def _merge_vols_in_mem(self, prob_container, label_container):
        max_prob_idx = np.argmax(prob_container, axis=0)
        max_prob_idx = max_prob_idx[np.newaxis, :, :, :]
        prob_container[0] = np.squeeze(
            np.take_along_axis(prob_container, max_prob_idx, axis=0)
        )
        label_container[0] = np.squeeze(
            np.take_along_axis(label_container, max_prob_idx, axis=0)
        )

    def _predict_12_ways_max_probs(self, data_vol):
        shape_tup = data_vol.shape
        logging.info("Creating empty data volumes in RAM to combine 12 way prediction.")
        label_container = np.empty((2, *shape_tup), dtype=np.uint8)
        prob_container = np.empty((2, *shape_tup), dtype=np.float16)
        label_container[0], prob_container[0] = self._predict_3_ways_max_probs(data_vol)
        for k in range(1, 4):
            logging.info(f"Rotating volume {k * 90} degrees")
            data_vol = np.rot90(data_vol)
            labels, probs = self._predict_3_ways_max_probs(data_vol)
            label_container[1] = np.rot90(labels, -k)
            prob_container[1] = np.rot90(probs, -k)
            logging.info(
                f"Merging rot {k * 90} deg volume with rot {(k-1) * 90} deg volume."
            )
            self._merge_vols_in_mem(prob_container, label_container)
        return label_container[0], prob_container[0]

    def _predict_single_axis_to_one_hot(self, data_vol, axis=Axis.Z):
        prediction, _ = self._predict_single_axis(data_vol, axis=axis)
        return utils.one_hot_encode_array(prediction, self.num_labels)

    def _predict_3_ways_one_hot(self, data_vol):
        one_hot_out = self._predict_single_axis_to_one_hot(data_vol)
        one_hot_out += self._predict_single_axis_to_one_hot(data_vol, Axis.Y)
        one_hot_out += self._predict_single_axis_to_one_hot(data_vol, Axis.X)
        return one_hot_out

    def _predict_12_ways_one_hot(self, data_vol):
        one_hot_out = self._predict_3_ways_one_hot(data_vol)
        for k in range(1, 4):
            logging.info(f"Rotating volume {k * 90} degrees")
            data_vol = np.rot90(data_vol)
            one_hot_out += np.rot90(
                self._predict_3_ways_one_hot(data_vol), -k, axes=(-3, -2)
            )
        return one_hot_out



#!/usr/bin/env python

import logging
import sys
from datetime import date
from pathlib import Path

import volume_segmantics.utilities.config as cfg
from volume_segmantics.data import (TrainingDataSlicer,
                                    get_settings_data)
from volume_segmantics.model import VolSeg2dTrainer
from volume_segmantics.utilities import get_2d_training_parser


def main():
    logging.basicConfig(
        level=logging.INFO, format=cfg.LOGGING_FMT, datefmt=cfg.LOGGING_DATE_FMT
    )
    # Parse args and check correct number of volumes given
    parser = get_2d_training_parser()
    args = parser.parse_args()
    data_vols = getattr(args, cfg.TRAIN_DATA_ARG)
    label_vols = getattr(args, cfg.LABEL_DATA_ARG)
    root_path = Path(getattr(args, cfg.DATA_DIR_ARG)).resolve()
    if len(data_vols) != len(label_vols):
        logging.error(
            "Number of data volumes and number of label volumes must be equal!"
        )
        sys.exit(1)
    # Create the settings object
    settings_path = Path(root_path, cfg.SETTINGS_DIR, cfg.TRAIN_SETTINGS_FN)
    settings = get_settings_data(settings_path)
    data_im_out_dir = root_path / settings.data_im_dirname  # dir for data imgs
    seg_im_out_dir = root_path / settings.seg_im_out_dirname  # dir for seg imgs
    # Keep track of the number of labels
    max_label_no = 0
    label_codes = None
    # Set up the DataSlicer and slice the data volumes into image files
    for count, (data_vol_path, label_vol_path) in enumerate(zip(data_vols, label_vols)):
        slicer = TrainingDataSlicer(data_vol_path, label_vol_path, settings)
        data_prefix, label_prefix = f"data{count}", f"seg{count}"
        slicer.output_data_slices(data_im_out_dir, data_prefix)
        slicer.output_label_slices(seg_im_out_dir, label_prefix)
        if slicer.num_seg_classes > max_label_no:
            max_label_no = slicer.num_seg_classes
            label_codes = slicer.codes
    assert label_codes is not None
    # Set up the 2dTrainer
    trainer = VolSeg2dTrainer(data_im_out_dir, seg_im_out_dir, max_label_no, settings)
    # Train the model, first frozen, then unfrozen
    num_cyc_frozen = settings.num_cyc_frozen
    num_cyc_unfrozen = settings.num_cyc_unfrozen
    model_type = settings.model["type"].name
    model_fn = f"{date.today()}_{model_type}_{settings.model_output_fn}.pytorch"
    model_out = Path(root_path, model_fn)
    if num_cyc_frozen > 0:
        trainer.train_model(
            model_out, num_cyc_frozen, settings.patience, create=True, frozen=True
        )
    if num_cyc_unfrozen > 0 and num_cyc_frozen > 0:
        trainer.train_model(
            model_out, num_cyc_unfrozen, settings.patience, create=False, frozen=False
        )
    elif num_cyc_unfrozen > 0 and num_cyc_frozen == 0:
        trainer.train_model(
            model_out, num_cyc_unfrozen, settings.patience, create=True, frozen=False
        )
    trainer.output_loss_fig(model_out)
    trainer.output_prediction_figure(model_out)
    # Clean up all the saved slices
    slicer.clean_up_slices()


if __name__ == "__main__":
    main()



"""
    This file is adapted from Adrian Wolny's pytorch-3dunet repository found at
    https://github.com/wolny/pytorch-3dunet
"""

import importlib

import torch
from torch import optim


def expand_as_one_hot(input, C, ignore_index=None):
    """
    Converts NxSPATIAL label image to NxCxSPATIAL, where each label gets converted to its corresponding one-hot vector.
    It is assumed that the batch dimension is present.
    Args:
        input (torch.Tensor): 3D/4D input image
        C (int): number of channels/labels
        ignore_index (int): ignore index to be kept during the expansion
    Returns:
        4D/5D output torch.Tensor (NxCxSPATIAL)
    """
    assert input.dim() == 4

    # expand the input tensor to Nx1xSPATIAL before scattering
    input = input.unsqueeze(1)
    # create output tensor shape (NxCxSPATIAL)
    shape = list(input.size())
    shape[1] = C

    if ignore_index is not None:
        # create ignore_index mask for the result
        mask = input.expand(shape) == ignore_index
        # clone the src tensor and zero out ignore_index in the input
        input = input.clone()
        input[input == ignore_index] = 0
        # scatter to get the one-hot tensor
        result = torch.zeros(shape).to(input.device).scatter_(1, input, 1)
        # bring back the ignore_index in the result
        result[mask] = ignore_index
        return result
    else:
        # scatter to get the one-hot tensor
        return torch.zeros(shape).to(input.device).scatter_(1, input, 1)


def convert_to_numpy(*inputs):
    """
    Coverts input tensors to numpy ndarrays

    Args:
        inputs (iteable of torch.Tensor): torch tensor

    Returns:
        tuple of ndarrays
    """

    def _to_numpy(i):
        assert isinstance(i, torch.Tensor), "Expected input to be torch.Tensor"
        return i.detach().cpu().numpy()

    return (_to_numpy(i) for i in inputs)



import argparse
from pathlib import Path

import volume_segmantics.utilities.config as cfg


def CheckExt(choices):
    """Wrapper to return the class"""

    class Act(argparse.Action):
        """Class to allow checking of filename extensions in argparse. Also
        checks whether file exists. Adapted from
        https://stackoverflow.com/questions/15203829/python-argparse-file-extension-checking
        """

        def __call__(self, parser, namespace, fnames, option_string=None):
            # Modified to take in a list of filenames
            if isinstance(fnames, list):
                for fname in fnames:
                    self.check_path(parser, fname)
            else:
                self.check_path(parser, fnames)

            # If all okay, set attribute
            setattr(namespace, self.dest, fnames)

        def check_path(self, parser, fname):
            fname = Path(fname)
            ext = fname.suffix
            if ext not in choices:
                parser.error(f"Wrong filetype: file {fname} doesn't end with {choices}")
                # Check that file exists
            if not fname.is_file():
                parser.error(f"The file {str(fname)} does not appear to exist.")

    return Act


def get_2d_training_parser() -> argparse.ArgumentParser:
    """Argument parser for scripts that train a 2d network on a 3d volume.

    Returns:
        argparse.ArgumentParser: An argument parser with the appropriate
        command line args contained within.
    """
    parser = argparse.ArgumentParser(
        usage="%(prog)s --data <path(s)/to/data/file(s)> --labels <path(s)/to/segmentation/file(s)> --data_dir path/to/data_directory",
        description="Train a 2d model on the 3d data and corresponding"
        " segmentation provided in the files.",
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"{parser.prog} version 1.0.0"
    )
    parser.add_argument(
        "--" + cfg.TRAIN_DATA_ARG,
        metavar="Path(s) to training image data volume(s)",
        type=str,
        action=CheckExt(cfg.TRAIN_DATA_EXT),
        nargs="+",
        required=True,
        help="the path(s) to file(s) containing the imaging data volume for training",
    )
    parser.add_argument(
        "--" + cfg.LABEL_DATA_ARG,
        metavar="Path(s) to label volume(s)",
        type=str,
        action=CheckExt(cfg.LABEL_DATA_EXT),
        nargs="+",
        required=True,
        help="the path(s) to file(s) containing a segmented volume for training",
    )
    parser.add_argument(
        "--" + cfg.DATA_DIR_ARG,
        metavar="Path to settings and output directory (optional)",
        type=str,
        nargs="?",
        default=Path.cwd(),
        help='path to a directory containing the "volseg-settings", data will be also be output to this location',
    )
    return parser


def get_2d_prediction_parser() -> argparse.ArgumentParser:
    """Argument parser for scripts that use a 2d network to predict segmenation for a 3d volume.

    Returns:
        argparse.ArgumentParser: An argument parser with the appropriate
        command line args contained within.
    """
    parser = argparse.ArgumentParser(
        usage="%(prog)s path/to/model/file.zip path/to/data/file [path/to/data_directory]",
        description="Predict segmentation of a 3d data volume using the 2d"
        " model provided.",
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"{parser.prog} version 1.0.0"
    )
    parser.add_argument(
        cfg.MODEL_PTH_ARG,
        metavar="Model file path",
        type=str,
        action=CheckExt(cfg.MODEL_DATA_EXT),
        help="the path to a zip file containing the model weights.",
    )
    parser.add_argument(
        cfg.PREDICT_DATA_ARG,
        metavar="Path to prediction data volume",
        type=str,
        action=CheckExt(cfg.PREDICT_DATA_EXT),
        help="the path to an HDF5 file containing the imaging data to segment",
    )
    parser.add_argument(
        "--" + cfg.DATA_DIR_ARG,
        metavar="Path to settings and output directory (optional)",
        type=str,
        nargs="?",
        default=Path.cwd(),
        help='path to a directory containing the "volseg-settings", data will be also be output to this location',
    )
    return parser



import logging

import numpy as np
import torch


class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience."""

    def __init__(self, patience=7, verbose=False, delta=0, path='checkpoint.pt', model_dict={}, best_score=None):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement. 
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
            path (str): Path for the checkpoint to be saved to.
                            Default: 'checkpoint.pt'
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = best_score
        self.val_loss_min = np.inf if best_score is None else best_score * -1
        self.early_stop = False
        self.delta = delta
        self.path = path
        self.model_struc_dict = model_dict # Dictionary with parameters controlling architecture

    def __call__(self, val_loss, model, optimizer, label_codes):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model, optimizer, label_codes)
        elif score < self.best_score + self.delta:
            self.counter += 1
            logging.info(
                f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model, optimizer, label_codes)
            self.counter = 0

    def save_checkpoint(self, val_loss, model, optimizer, label_codes):
        '''Saves model when validation loss decrease.'''
        if self.verbose:
            logging.info(
                f'Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ...')
        model_dict = {
            "model_state_dict": model.state_dict(),
            "model_struc_dict": self.model_struc_dict,
            "optimizer_state_dict": optimizer.state_dict(),
            "loss_val": val_loss,
            "label_codes": label_codes,
        }        
        torch.save(model_dict, self.path)
        self.val_loss_min = val_loss



import logging
import sys
from pathlib import Path
from types import SimpleNamespace
from typing import Union

import yaml


def get_settings_data(data: Union[Path, dict, None]) -> SimpleNamespace:
    """Given a path to a YAML file or a dictionary object, returns a 
    simple namespace object holding settings data. If the data is None, 
    an empty namespace is returned.
    """
    if data is None:
        return SimpleNamespace()
    elif isinstance(data, Path):
        logging.info(f"Loading settings from {data}")
        if data.exists():
            with open(data, "r") as stream:
                settings_dict = yaml.safe_load(stream)
            return SimpleNamespace(**settings_dict)
        else:
            logging.error("Couldn't find settings file... Exiting!")
            sys.exit(1)
    elif isinstance(data, dict):
        return SimpleNamespace(**data)



#!/usr/bin/env python

import logging
import warnings
from datetime import date
from pathlib import Path

import volume_segmantics.utilities.config as cfg
from volume_segmantics.data import get_settings_data
from volume_segmantics.model import VolSeg2DPredictionManager
from volume_segmantics.utilities import get_2d_prediction_parser

warnings.filterwarnings("ignore", category=UserWarning)


def create_output_path(root_path, data_vol_path):
    pred_out_fn = f"{date.today()}_{data_vol_path.stem}_2d_model_vol_pred.h5"
    return Path(root_path, pred_out_fn)


def main():
    logging.basicConfig(
        level=logging.INFO, format=cfg.LOGGING_FMT, datefmt=cfg.LOGGING_DATE_FMT
    )
    # Parse Args
    parser = get_2d_prediction_parser()
    args = parser.parse_args()
    # Define paths
    root_path = Path(getattr(args, cfg.DATA_DIR_ARG)).resolve()
    settings_path = Path(root_path, cfg.SETTINGS_DIR, cfg.PREDICTION_SETTINGS_FN)
    model_file_path = getattr(args, cfg.MODEL_PTH_ARG)
    data_vol_path = Path(getattr(args, cfg.PREDICT_DATA_ARG))
    output_path = create_output_path(root_path, data_vol_path)
    # Get settings object
    settings = get_settings_data(settings_path)
    # Create prediction manager and predict
    pred_manager = VolSeg2DPredictionManager(model_file_path, data_vol_path, settings)
    pred_manager.predict_volume_to_path(output_path)


if __name__ == "__main__":
    main()



"""Data to be shared across files.
"""
# Parser strings
TRAIN_DATA_ARG = "data"
LABEL_DATA_ARG = "labels"
MODEL_PTH_ARG = "model"
PREDICT_DATA_ARG = "data"
DATA_DIR_ARG = "data_dir"
# File extensions
TIFF_SUFFIXES = {".tiff", ".tif"}
HDF5_SUFFIXES = {".h5", ".hdf5", ".nxs"}
TRAIN_DATA_EXT = {*HDF5_SUFFIXES, *TIFF_SUFFIXES}
LABEL_DATA_EXT = {*HDF5_SUFFIXES, *TIFF_SUFFIXES}
MODEL_DATA_EXT = {".pytorch", ".pth"}
PREDICT_DATA_EXT = {*HDF5_SUFFIXES, *TIFF_SUFFIXES}
# TODO Required settings - check required keys are in settings files
# Logging format
LOGGING_FMT = "%(asctime)s - %(levelname)s - %(message)s"
LOGGING_DATE_FMT = "%d-%b-%y %H:%M:%S"
# Settings yaml file locations
SETTINGS_DIR = "volseg-settings"
TRAIN_SETTINGS_FN = "2d_model_train_settings.yaml"
PREDICTION_SETTINGS_FN = "2d_model_predict_settings.yaml"

TQDM_BAR_FORMAT = "{l_bar}{bar: 30}{r_bar}{bar: -30b}"  # tqdm progress bar format

HDF5_COMPRESSION = "gzip"

BIG_CUDA_THRESHOLD = 8 # GPU Memory (GB), above this value batch size is increased
BIG_CUDA_TRAIN_BATCH = 12 # Size of training batch on big GPU
BIG_CUDA_PRED_BATCH = 4 # Size of prediction batch on big GPU
SMALL_CUDA_BATCH = 2 # Size of batch on small GPU
NUM_WORKERS = 4 # Number of parallel workers for training/validation dataloaders
PIN_CUDA_MEMORY = True # Whether to pin CUDA memory for faster data transfer
IM_SIZE_DIVISOR = 32 # Image dimensions need to be a multiple of this value
MODEL_INPUT_CHANNELS = 1 # Use 1 for grayscale input images 

DEFAULT_MIN_LR = 0.00075 # Learning rate to return if LR finder fails
LR_DIVISOR = 3 # Divide the automatically calculated learning rate (min gradient) by this magic number

IMAGENET_MEAN = 0.449 # Mean value for single channel imagnet normalisation
IMAGENET_STD = 0.226 # Standard deviation for single channel imagenet normalisation
 


import logging
from pathlib import Path
from types import SimpleNamespace
from typing import Union

import numpy as np
import volume_segmantics.utilities.base_data_utils as utils


class BaseDataManager:
    def __init__(
        self, data_vol: Union[Path, str, np.ndarray], settings: SimpleNamespace
    ) -> None:
        self.data_vol_shape = None
        self.data_mean = None
        self.data_vol_path = utils.setup_path_if_exists(data_vol)
        self.settings = settings
        self.st_dev_factor = settings.st_dev_factor
        self.downsample = settings.downsample
        if self.data_vol_path is not None:
            self.data_vol, self.input_data_chunking = utils.get_numpy_from_path(
                self.data_vol_path, internal_path=settings.data_hdf5_path
            )
        elif isinstance(data_vol, np.ndarray):
            self.data_vol = data_vol
            self.input_data_chunking = True
        self._preprocess_data()

    def _preprocess_data(self):
        if self.downsample:
            self.data_vol = utils.downsample_data(self.data_vol)
        self.data_vol_shape = self.data_vol.shape
        logging.info("Calculating mean of data...")
        self.data_mean = np.nanmean(self.data_vol)
        logging.info(f"Mean value: {self.data_mean}")
        if self.settings.clip_data:
            self.data_vol = utils.clip_to_uint8(
                self.data_vol, self.data_mean, self.st_dev_factor
            )
        if np.isnan(self.data_vol).any():
            logging.info(f"Replacing NaN values.")
            self.data_vol = np.nan_to_num(self.data_vol, copy=False)



"""
Provides functions to enable image augmentations for training and prediction.
"""

import math

import albumentations as A
from albumentations.pytorch.transforms import ToTensorV2
import volume_segmantics.utilities.config as cfg


def get_train_preprocess_augs(img_size: int) -> A.core.composition.Compose:
    """Returns the augmentations required to pad images to the correct square size
    for training a network.

    Args:
        img_size (int): Length of square image edge.

    Returns:
        A.core.composition.Compose: An augmentation to resize the image if needed.
    """
    return A.Compose(
        [
            A.LongestMaxSize(max_size=img_size, p=1.0),
            A.PadIfNeeded(min_height=img_size, min_width=img_size, p=1.0),
        ]
    )


def get_padded_dimension(dimension: int) -> int:
    """Returns the closest image dimension that can be divided by the divisor
    specified in the config.

    Args:
        dimension (int): input dimension.

    Returns:
        int: Dimension that can divided by divisor
    """
    image_divisor = cfg.IM_SIZE_DIVISOR
    if dimension % image_divisor == 0:
        return dimension
    return (math.floor(dimension / image_divisor) + 1) * image_divisor


def get_pred_preprocess_augs(
    img_size_y: int, img_size_x: int
) -> A.core.composition.Compose:
    """Returns the augmentations required to pad images to the correct size for
    prediction

    Args:
        img_size_y (int): Image size in y
        img_size_x (int): Image size in x

    Returns:
        A.core.composition.Compose: An augmentation to resize the image if needed.
    """
    padded_y_dim = get_padded_dimension(img_size_y)
    padded_x_dim = get_padded_dimension(img_size_x)
    return A.Compose(
        [
            A.PadIfNeeded(min_height=padded_y_dim, min_width=padded_x_dim, p=1.0),
        ]
    )


def get_train_augs(img_size: int) -> A.core.composition.Compose:
    """Returns the augmentations used for training a network.

    Args:
        img_size (int): The square image size required.

    Returns:
        A.core.composition.Compose: Augmentations for training.
    """
    return A.Compose(
        [
            A.RandomSizedCrop(
                min_max_height=(img_size // 2, img_size),
                height=img_size,
                width=img_size,
                p=0.5,
            ),
            A.VerticalFlip(p=0.5),
            A.RandomRotate90(p=0.5),
            A.Transpose(p=0.5),
            A.OneOf(
                [
                    A.ElasticTransform(
                        alpha=120, sigma=120 * 0.07, alpha_affine=120 * 0.04, p=0.5
                    ),
                    A.GridDistortion(p=0.5),
                    A.OpticalDistortion(distort_limit=1, shift_limit=0.5, p=0.5),
                ],
                p=0.5,
            ),
            A.CLAHE(p=0.5),
            A.OneOf([A.RandomBrightnessContrast(p=0.5), A.RandomGamma(p=0.5)], p=0.5),
        ]
    )


def get_postprocess_augs() -> A.core.composition.Compose:
    """Returns the final augmentations applied to the images.

    Returns:
        A.core.composition.Compose: Postprocessing augmentations.
    """
    return A.Compose([ToTensorV2()])



import logging
from pathlib import Path
from typing import Tuple

import segmentation_models_pytorch as smp
import torch
import volume_segmantics.utilities.base_data_utils as utils


def create_model_on_device(device_num: int, model_struc_dict: dict) -> torch.nn.Module:

    struct_dict_copy = model_struc_dict.copy()
    model_type = struct_dict_copy.pop("type")
    
    if model_type == utils.ModelType.U_NET:
        model = smp.Unet(**struct_dict_copy)
        logging.info(f"Sending the U-Net model to device {device_num}")
    elif model_type == utils.ModelType.U_NET_PLUS_PLUS:
        model = smp.UnetPlusPlus(**struct_dict_copy)
        logging.info(f"Sending the U-Net++ model to device {device_num}")
    elif model_type == utils.ModelType.FPN:
        model = smp.FPN(**struct_dict_copy)
        logging.info(f"Sending the FPN model to device {device_num}")
    elif model_type == utils.ModelType.DEEPLABV3:
        model = smp.DeepLabV3(**struct_dict_copy)
        logging.info(f"Sending the DeepLabV3 model to device {device_num}")
    elif model_type == utils.ModelType.DEEPLABV3_PLUS:
        model = smp.DeepLabV3Plus(**struct_dict_copy)
        logging.info(f"Sending the DeepLabV3+ model to device {device_num}")
    elif model_type == utils.ModelType.MA_NET:
        model = smp.MAnet(**struct_dict_copy)
        logging.info(f"Sending the MA-Net model to device {device_num}")
    elif model_type == utils.ModelType.LINKNET:
        model = smp.Linknet(**struct_dict_copy)
        logging.info(f"Sending the Linknet model to device {device_num}")
    elif model_type == utils.ModelType.PAN:
        model = smp.PAN(**struct_dict_copy)
        logging.info(f"Sending the PAN model to device {device_num}")
    return model.to(device_num)


def create_model_from_file(
    weights_fn: Path, gpu: bool = True, device_num: int = 0,
) -> Tuple[torch.nn.Module, int, dict]:
    """Creates and returns a model and the number of segmentation labels
    that are predicted by the model."""
    if gpu:
        map_location = f"cuda:{device_num}"
    else:
        map_location = "cpu"
    weights_fn = weights_fn.resolve()
    logging.info("Loading model dictionary from file.")
    model_dict = torch.load(weights_fn, map_location=map_location)
    model = create_model_on_device(device_num, model_dict["model_struc_dict"])
    logging.info("Loading in the saved weights.")
    model.load_state_dict(model_dict["model_state_dict"])
    return model, model_dict["model_struc_dict"]["classes"], model_dict["label_codes"]



from pathlib import Path
from types import SimpleNamespace
from typing import Tuple

import numpy as np
import torch
import volume_segmantics.utilities.base_data_utils as utils
import volume_segmantics.utilities.config as cfg
from torch.utils.data import DataLoader, Subset
from volume_segmantics.data.datasets import (get_2d_prediction_dataset,
                                             get_2d_training_dataset,
                                             get_2d_validation_dataset)


def get_2d_training_dataloaders(
    image_dir: Path, label_dir: Path, settings: SimpleNamespace
) -> Tuple[DataLoader, DataLoader]:
    """Returns 2d training and validation dataloaders with indices split at random
    according to the percentage split specified in settings.

    Args:
        image_dir (Path): Directory of data images
        label_dir (Path): Directory of label images
        settings (SimpleNamespace): Settings object

    Returns:
        Tuple[DataLoader, DataLoader]: 2d training and validation dataloaders
    """

    training_set_prop = settings.training_set_proportion
    batch_size = utils.get_batch_size(settings)

    full_training_dset = get_2d_training_dataset(image_dir, label_dir, settings)
    full_validation_dset = get_2d_validation_dataset(image_dir, label_dir, settings)
    # split the dataset in train and test set
    dset_length = len(full_training_dset)
    indices = torch.randperm(dset_length).tolist()
    train_idx, validate_idx = np.split(indices, [int(dset_length * training_set_prop)])
    training_dataset = Subset(full_training_dset, train_idx)
    validation_dataset = Subset(full_validation_dset, validate_idx)

    training_dataloader = DataLoader(
        training_dataset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=cfg.NUM_WORKERS,
        pin_memory=cfg.PIN_CUDA_MEMORY,
        drop_last=True,
    )
    validation_dataloader = DataLoader(
        validation_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=cfg.NUM_WORKERS,
        pin_memory=cfg.PIN_CUDA_MEMORY,
    )
    return training_dataloader, validation_dataloader


def get_2d_prediction_dataloader(
    data_vol: np.array, settings: SimpleNamespace
) -> DataLoader:
    pred_dataset = get_2d_prediction_dataset(data_vol)
    batch_size = utils.get_batch_size(settings, prediction=True)
    return DataLoader(
        pred_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=0,  # Set to 0 for prediction
        pin_memory=cfg.PIN_CUDA_MEMORY,
    )



__all__ = ["VolSeg2dTrainer", "VolSeg2DPredictionManager"]

from volume_segmantics.model.operations.vol_seg_2d_trainer import VolSeg2dTrainer
from volume_segmantics.model.operations.vol_seg_prediction_manager import (
    VolSeg2DPredictionManager,
)



from pathlib import Path
from types import SimpleNamespace
from typing import Union

import numpy as np

import volume_segmantics.utilities.base_data_utils as utils
from volume_segmantics.data.base_data_manager import BaseDataManager
from volume_segmantics.model.operations.vol_seg_2d_predictor import VolSeg2dPredictor


class VolSeg2DPredictionManager(BaseDataManager):
    """Class that manages prediction of data volumes to disk using a
    2d deep learning network.
    """

    def __init__(
        self,
        model_file_path: str,
        data_vol: Union[str, np.ndarray],
        settings: SimpleNamespace,
    ) -> None:
        """Inits VolSeg2DPredictionManager.

        Args:
            model_file_path (str): String of filepath to trained model to use for prediction.
            data_vol (Union[str, np.ndarray]): String of filepath to data volume or numpy array of data to predict segmentation of.
            settings (SimpleNamespace): A prediction settings object.
        """
        super().__init__(data_vol, settings)
        self.predictor = VolSeg2dPredictor(model_file_path, settings)
        self.settings = settings

    def get_label_codes(self) -> dict:
        """Returns a dictionary of label codes, retrieved from the saved model.

        Returns:
            dict: Label codes. These provide information on the labels that were used
            when training the model along with any associated metadata.
        """
        return self.predictor.label_codes

    def predict_volume_to_path(
        self, output_path: Union[Path, None], quality: Union[utils.Quality, None] = None
    ) -> np.ndarray:
        """Method which triggers prediction of a 3D segmentation to disk at a specified quality.

        Here 'quality' refers to the number of axes/rotations that the segmentation is predicted
        in. e.g. Low quality, single axis (x, y) prediction; medium quality, three axis (x, y),
        (x, z), (y, z) prediction; high quality 12 way (3 axis and 4 rotations) prediction.
        Multi-axis predictions are combined into a final output volume by using maximum probabilities.

        Args:
            output_path (Union[Path, None]): Path to predict volume to.
            quality (Union[utils.Quality, None], optional): A quality to predict the segmentation to. Defaults to None.

        Returns:
            np.ndarray: _description_
        """
        probs = None
        one_hot = self.settings.one_hot
        preferred_axis = utils.get_prediction_axis(
            self.settings
        )  # Specify single axis for prediction
        if quality is None:
            quality = utils.get_prediction_quality(self.settings)
        if quality == utils.Quality.LOW:
            if one_hot:
                prediction = self.predictor._predict_single_axis_to_one_hot(
                    self.data_vol, axis=preferred_axis
                )
            else:
                prediction, probs = self.predictor._predict_single_axis(
                    self.data_vol, axis=preferred_axis
                )
        if quality == utils.Quality.MEDIUM:
            if one_hot:
                prediction = self.predictor._predict_3_ways_one_hot(self.data_vol)
            else:
                prediction, probs = self.predictor._predict_3_ways_max_probs(
                    self.data_vol
                )
        if quality == utils.Quality.HIGH:
            if one_hot:
                prediction = self.predictor._predict_12_ways_one_hot(self.data_vol)
            else:
                prediction, probs = self.predictor._predict_12_ways_max_probs(
                    self.data_vol
                )
        if output_path is not None:
            utils.save_data_to_hdf5(
                prediction, output_path, chunking=self.input_data_chunking
            )
            if probs is not None and self.settings.output_probs:
                utils.save_data_to_hdf5(
                    probs,
                    f"{output_path.parent / output_path.stem}_probs.h5",
                    chunking=self.input_data_chunking,
                )
        return prediction



"""
    This file is adapted from Adrian Wolny's pytorch-3dunet repository found at
    https://github.com/wolny/pytorch-3dunet
"""

import importlib

import numpy as np
import torch

from skimage.metrics import peak_signal_noise_ratio, mean_squared_error

from volume_segmantics.data.pytorch3dunet_losses import compute_per_channel_dice
from volume_segmantics.utilities.pytorch3dunet_utils import expand_as_one_hot, convert_to_numpy


class DiceCoefficient:
    """Computes Dice Coefficient.
    Generalized to multiple channels by computing per-channel Dice Score
    (as described in https://arxiv.org/pdf/1707.03237.pdf) and theTn simply taking the average.
    Input is expected to be probabilities instead of logits.
    This metric is mostly useful when channels contain the same semantic class (e.g. affinities computed with different offsets).
    DO NOT USE this metric when training with DiceLoss, otherwise the results will be biased towards the loss.
    """

    def __init__(self, epsilon=1e-6, **kwargs):
        self.epsilon = epsilon

    def __call__(self, input, target):
        # Average across channels in order to get the final score
        return torch.mean(compute_per_channel_dice(input, target, epsilon=self.epsilon))


class MeanIoU:
    """
    Computes IoU for each class separately and then averages over all classes.
    """

    def __init__(self, skip_channels=(), ignore_index=None, **kwargs):
        """
        :param skip_channels: list/tuple of channels to be ignored from the IoU computation
        :param ignore_index: id of the label to be ignored from IoU computation
        """
        self.ignore_index = ignore_index
        self.skip_channels = skip_channels

    def __call__(self, input, target):
        """
        :param input: 5D probability maps torch float tensor (NxCxDxHxW)
        :param target: 4D or 5D ground truth torch tensor. 4D (NxDxHxW) tensor will be expanded to 5D as one-hot
        :return: intersection over union averaged over all channels
        """
        assert input.dim() == 5

        n_classes = input.size()[1]

        if target.dim() == 4:
            target = expand_as_one_hot(target, C=n_classes, ignore_index=self.ignore_index)

        assert input.size() == target.size()

        per_batch_iou = []
        for _input, _target in zip(input, target):
            binary_prediction = self._binarize_predictions(_input, n_classes)

            if self.ignore_index is not None:
                # zero out ignore_index
                mask = _target == self.ignore_index
                binary_prediction[mask] = 0
                _target[mask] = 0

            # convert to uint8 just in case
            binary_prediction = binary_prediction.byte()
            _target = _target.byte()

            per_channel_iou = []
            for c in range(n_classes):
                if c in self.skip_channels:
                    continue

                per_channel_iou.append(self._jaccard_index(binary_prediction[c], _target[c]))

            assert per_channel_iou, "All channels were ignored from the computation"
            mean_iou = torch.mean(torch.tensor(per_channel_iou))
            per_batch_iou.append(mean_iou)

        return torch.mean(torch.tensor(per_batch_iou))

    def _binarize_predictions(self, input, n_classes):
        """
        Puts 1 for the class/channel with the highest probability and 0 in other channels. Returns byte tensor of the
        same size as the input tensor.
        """
        if n_classes == 1:
            # for single channel input just threshold the probability map
            result = input > 0.5
            return result.long()

        _, max_index = torch.max(input, dim=0, keepdim=True)
        return torch.zeros_like(input, dtype=torch.uint8).scatter_(0, max_index, 1)

    def _jaccard_index(self, prediction, target):
        """
        Computes IoU for a given target and prediction tensors
        """
        return torch.sum(prediction & target).float() / torch.clamp(torch.sum(prediction | target).float(), min=1e-8)


class PSNR:
    """
    Computes Peak Signal to Noise Ratio. Use e.g. as an eval metric for denoising task
    """

    def __init__(self, **kwargs):
        pass

    def __call__(self, input, target):
        input, target = convert_to_numpy(input, target)
        return peak_signal_noise_ratio(target, input)


class MSE:
    """
    Computes MSE between input and target
    """

    def __init__(self, **kwargs):
        pass

    def __call__(self, input, target):
        input, target = convert_to_numpy(input, target)
        return mean_squared_error(input, target)


def get_evaluation_metric(config):
    """
    Returns the evaluation metric function based on provided configuration
    :param config: (dict) a top level configuration object containing the 'eval_metric' key
    :return: an instance of the evaluation metric
    """

    def _metric_class(class_name):
        m = importlib.import_module('pytorch3dunet.unet3d.metrics')
        clazz = getattr(m, class_name)
        return clazz

    assert 'eval_metric' in config, 'Could not find evaluation metric configuration'
    metric_config = config['eval_metric']
    metric_class = _metric_class(metric_config['name'])
    return metric_class(**metric_config)



import re
from pathlib import Path
from types import SimpleNamespace

import cv2
import numpy as np
import volume_segmantics.data.augmentations as augs
import volume_segmantics.utilities.config as cfg
from torch.utils.data import Dataset as BaseDataset


class VolSeg2dDataset(BaseDataset):
    """Read images, apply augmentation and preprocessing transformations.

    Args:
        images_dir (pathlib.Path): path to images folder
        masks_dir (pathlib.Path): path to segmentation masks folder
        preprocessing (albumentations.Compose): data pre-processing
            (e.g. padding, resizing)
        augmentation (albumentations.Compose): data transformation pipeline
            (e.g. flip, scale, contrast adjustments)
        imagenet_norm (bool): Whether to normalise according to imagenet stats
        postprocessing (albumentations.Compose): data post-processing
            (e.g. Convert to Tensor)


    """

    imagenet_mean = cfg.IMAGENET_MEAN
    imagenet_std = cfg.IMAGENET_STD

    def __init__(
        self,
        images_dir,
        masks_dir,
        preprocessing=None,
        augmentation=None,
        imagenet_norm=True,
        postprocessing=None,
    ):

        self.images_fps = sorted(list(images_dir.glob("*.png")), key=self.natsort)
        self.masks_fps = sorted(list(masks_dir.glob("*.png")), key=self.natsort)
        self.augmentation = augmentation
        self.preprocessing = preprocessing
        self.imagenet_norm = imagenet_norm
        self.postprocessing = postprocessing

    def __getitem__(self, i):

        # read data
        image = cv2.imread(str(self.images_fps[i]), cv2.IMREAD_GRAYSCALE)
        mask = cv2.imread(str(self.masks_fps[i]), 0)

        # apply pre-processing
        if self.preprocessing:
            sample = self.preprocessing(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        # apply augmentations
        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        if self.imagenet_norm:
            if np.issubdtype(image.dtype, np.integer):
                # Convert to float
                image = image.astype(np.float32)
                image = image / 255
            image = image - self.imagenet_mean
            image = image / self.imagenet_std

        # apply post-processing
        if self.postprocessing:
            sample = self.postprocessing(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        return image, mask

    def __len__(self):
        return len(self.images_fps)

    @staticmethod
    def natsort(item):
        return [
            int(t) if t.isdigit() else t.lower() for t in re.split(r"(\d+)", str(item))
        ]


class VolSeg2dPredictionDataset(BaseDataset):
    """Splits 3D data volume into 2D images for inference.

    Args:
        images_dir (pathlib.Path): path to images folder
        masks_dir (pathlib.Path): path to segmentation masks folder
        preprocessing (albumentations.Compose): data pre-processing
            (e.g. padding, resizing)
        imagenet_norm (bool): Whether to normalise according to imagenet stats
        postprocessing (albumentations.Compose): data post-processing
            (e.g. Convert to Tensor)


    """

    imagenet_mean = cfg.IMAGENET_MEAN
    imagenet_std = cfg.IMAGENET_STD

    def __init__(
        self,
        data_vol,
        preprocessing=None,
        imagenet_norm=True,
        postprocessing=None,
    ):
        self.data_vol = data_vol
        self.preprocessing = preprocessing
        self.imagenet_norm = imagenet_norm
        self.postprocessing = postprocessing

    def __getitem__(self, i):

        image = self.data_vol[i]

        # apply pre-processing
        if self.preprocessing:
            sample = self.preprocessing(image=image)
            image = sample["image"]

        if self.imagenet_norm:
            if np.issubdtype(image.dtype, np.integer):
                # Convert to float
                image = image.astype(np.float32)
                image = image / 255
            image = image - self.imagenet_mean
            image = image / self.imagenet_std

        # apply post-processing
        if self.postprocessing:
            sample = self.postprocessing(image=image)
            image = sample["image"]

        return image

    def __len__(self):
        return self.data_vol.shape[0]


def get_2d_training_dataset(
    image_dir: Path, label_dir: Path, settings: SimpleNamespace
) -> VolSeg2dDataset:

    img_size = settings.image_size
    return VolSeg2dDataset(
        image_dir,
        label_dir,
        preprocessing=augs.get_train_preprocess_augs(img_size),
        augmentation=augs.get_train_augs(img_size),
        postprocessing=augs.get_postprocess_augs(),
    )


def get_2d_validation_dataset(
    image_dir: Path, label_dir: Path, settings: SimpleNamespace
) -> VolSeg2dDataset:

    img_size = settings.image_size
    return VolSeg2dDataset(
        image_dir,
        label_dir,
        preprocessing=augs.get_train_preprocess_augs(img_size),
        postprocessing=augs.get_postprocess_augs(),
    )


def get_2d_prediction_dataset(data_vol: np.array) -> VolSeg2dPredictionDataset:
    y_dim, x_dim = data_vol.shape[1:]
    return VolSeg2dPredictionDataset(
        data_vol,
        preprocessing=augs.get_pred_preprocess_augs(y_dim, x_dim),
        postprocessing=augs.get_postprocess_augs(),
    )



import logging
import pathlib
import sys
from enum import Enum
from itertools import chain, product
from types import SimpleNamespace
from typing import List, Union, Tuple

import h5py as h5
import imageio
import numpy as np
import torch
import torchvision.transforms.functional as F
from skimage.measure import block_reduce

import volume_segmantics.utilities.config as cfg
from types import SimpleNamespace
from pathlib import Path


class Quality(Enum):
    """An Enum that holds values describing the quality of segmentation
    prediction required.

    Refers to the number of axes/rotations that the segmentation is predicted
    in. e.g. Low quality, single axis (x, y) prediction; medium quality, three axis (x, y),
    (x, z), (y, z) prediction; high quality 12 way (3 axis and 4 rotations) prediction.
    """

    LOW = 1
    MEDIUM = 3
    HIGH = 12


class Axis(Enum):
    Z = 0
    Y = 1
    X = 2
    ALL = 4


class ModelType(Enum):
    U_NET = 1
    U_NET_PLUS_PLUS = 2
    FPN = 3
    DEEPLABV3 = 4
    DEEPLABV3_PLUS = 5
    MA_NET = 6
    LINKNET = 7
    PAN = 8


def create_enum_from_setting(setting_str, enum):
    if isinstance(setting_str, Enum):
        return setting_str
    try:
        output_enum = enum[setting_str.upper()]
    except KeyError:
        options = [k.name for k in enum]
        logging.error(
            f"{enum.__name__}: {setting_str} is not valid. Options are {options}."
        )
        sys.exit(1)
    return output_enum


def get_prediction_quality(settings: SimpleNamespace) -> Enum:
    pred_quality = create_enum_from_setting(settings.quality, Quality)
    return pred_quality


def get_model_type(settings: SimpleNamespace) -> Enum:
    model_type = create_enum_from_setting(settings.model["type"], ModelType)
    return model_type


def get_training_axis(settings: SimpleNamespace) -> Enum:
    try:
        axis_setting = settings.training_axes
    except AttributeError:
        axis_setting = "All"
    axis_enum = create_enum_from_setting(axis_setting, Axis)
    return axis_enum


def get_prediction_axis(settings: SimpleNamespace) -> Enum:
    try:
        axis_setting = settings.prediction_axis
    except AttributeError:
        axis_setting = "Z"
    axis_enum = create_enum_from_setting(axis_setting, Axis)
    return axis_enum


def setup_path_if_exists(input_param):
    if isinstance(input_param, str):
        return Path(input_param)
    elif isinstance(input_param, Path):
        return input_param
    else:
        return None


def get_batch_size(settings: SimpleNamespace, prediction: bool = False) -> int:

    cuda_device_num = settings.cuda_device
    total_gpu_mem = torch.cuda.get_device_properties(cuda_device_num).total_memory
    allocated_gpu_mem = torch.cuda.memory_allocated(cuda_device_num)
    free_gpu_mem = (total_gpu_mem - allocated_gpu_mem) / 1024**3

    if free_gpu_mem < cfg.BIG_CUDA_THRESHOLD:
        batch_size = cfg.SMALL_CUDA_BATCH
    elif not prediction:
        batch_size = cfg.BIG_CUDA_TRAIN_BATCH
    else:
        batch_size = cfg.BIG_CUDA_PRED_BATCH

    logging.info(
        f"Free GPU memory is {free_gpu_mem:0.2f} GB. Batch size will be "
        f"{batch_size}."
    )
    return batch_size


def crop_tensor_to_array(tensor: torch.Tensor, yx_dims: List[int]) -> np.array:
    if tensor.is_cuda:
        tensor = tensor.cpu()
    tensor = F.center_crop(tensor, yx_dims)
    return tensor.detach().numpy()


def rotate_array_to_axis(array: np.array, axis: Axis = Axis.Z) -> np.array:
    if axis == Axis.Z:
        return array
    if axis == Axis.Y:
        return array.swapaxes(0, 1)
    if axis == Axis.X:
        return array.swapaxes(0, 2)


def one_hot_encode_array(input_array: np.array, num_labels: int) -> np.array:
    """Modified from https://stackoverflow.com/questions/36960320/convert-a-2d-matrix-to-a-3d-one-hot-matrix-numpy"""
    ncols = num_labels
    out = np.zeros((ncols, input_array.size), dtype=np.uint8)
    out[input_array.ravel(), np.arange(input_array.size)] = 1
    out.shape = (ncols,) + input_array.shape
    return out


def prepare_training_batch(
    batch: "list[torch.Tensor]", device: int, num_labels: int
) -> "tuple[torch.Tensor, torch.Tensor]":
    inputs = batch[0].to(device)
    targets = batch[1].to(torch.int64)
    # One hot encode the channels
    targets = torch.nn.functional.one_hot(targets, num_classes=num_labels)
    targets = targets.permute((0, 3, 1, 2)).to(device, dtype=torch.uint8)
    return inputs, targets


def downsample_data(data, factor=2):
    logging.info(f"Downsampling data by a factor of {factor}.")
    return block_reduce(data, block_size=(factor, factor, factor), func=np.nanmean)


def numpy_from_tiff(path):
    """Returns a numpy array when given a path to an multipage TIFF file.

    Args:
        path(pathlib.Path): The path to the TIFF file.

    Returns:
        numpy.array: A numpy array object for the data stored in the TIFF file.
    """

    return imageio.volread(path)


def numpy_from_hdf5(path, hdf5_path="/data", nexus=False):
    """Returns a numpy array  and chunking info when given a path
    to an HDF5 file.

    The data is assumed to be found in '/data' in the file.

    Args:
        path(pathlib.Path): The path to the HDF5 file.
        hdf5_path (str): The internal HDF5 path to the data.

    Returns:
        tuple(numpy.array, tuple(int, int)) : A numpy array
        for the data and a tuple with the chunking size.
    """

    data_handle = h5.File(path, "r")
    if nexus:
        try:
            dataset = data_handle["processed/result/data"]
        except KeyError:
            logging.error(
                "NXS file: Couldn't find data at 'processed/result/data' trying another path."
            )
            try:
                dataset = data_handle["entry/final_result_tomo/data"]
            except KeyError:
                logging.error(
                    "NXS file: Could not find entry at entry/final_result_tomo/data, exiting!"
                )
                sys.exit(1)
    else:
        dataset = data_handle[hdf5_path]
    input_data_chunking = dataset.chunks
    return dataset[()], input_data_chunking


def get_numpy_from_path(
    path: pathlib.Path, internal_path: str = "/data"
) -> Tuple[np.array, Union[Tuple[int, int], bool]]:
    """Helper function that returns numpy array and chunking(if used)
    according to file extension.

    Args:
        path (pathlib.Path): Path to file
        internal_path (str, optional): Path inside HDF5 file. Defaults to "/data".

    Returns:
        Tuple[np.array, Union[Tuple[int, int], bool]]: Tuple with data array and
        either chunking tuple, or True.
    """
    if path.suffix in cfg.TIFF_SUFFIXES:
        return numpy_from_tiff(path), True
    elif path.suffix in cfg.HDF5_SUFFIXES:
        nexus = path.suffix == ".nxs"
        return numpy_from_hdf5(path, hdf5_path=internal_path, nexus=nexus)


def sequential_labels(unique_labels: np.array) -> bool:
    if not np.where(np.diff(unique_labels) != 1)[0].size:
        return True
    else:
        return False


def clip_to_uint8(data: np.array, data_mean: float, st_dev_factor: float) -> np.array:
    """Clips data to a certain number of st_devs of the mean and reduces
    bit depth to uint8.

    Args:
        data(np.array): The data to be processed.

    Returns:
        np.array: A unit8 data array.
    """
    logging.info("Clipping data and converting to uint8.")
    logging.info(f"Calculating standard deviation.")
    data_st_dev = np.nanstd(data)
    logging.info(f"Std dev: {data_st_dev}. Calculating stats.")
    num_vox = data.size
    lower_bound = data_mean - (data_st_dev * st_dev_factor)
    upper_bound = data_mean + (data_st_dev * st_dev_factor)
    with np.errstate(invalid="ignore"):
        gt_ub = (data > upper_bound).sum()
        lt_lb = (data < lower_bound).sum()
    logging.info(f"Lower bound: {lower_bound}, upper bound: {upper_bound}")
    logging.info(
        f"Number of voxels above upper bound to be clipped {gt_ub} - percentage {gt_ub/num_vox * 100:.3f}%"
    )
    logging.info(
        f"Number of voxels below lower bound to be clipped {lt_lb} - percentage {lt_lb/num_vox * 100:.3f}%"
    )
    if np.isnan(data).any():
        logging.info(f"Replacing NaN values.")
        data = np.nan_to_num(data, copy=False, nan=data_mean)
    logging.info("Rescaling intensities.")
    if np.issubdtype(data.dtype, np.integer):
        logging.info(
            "Data is already in integer dtype, converting to float for rescaling."
        )
        data = data.astype(float)
    data = np.clip(data, lower_bound, upper_bound, out=data)
    data = np.subtract(data, lower_bound, out=data)
    data = np.divide(data, (upper_bound - lower_bound), out=data)
    # data = (data - lower_bound) / (upper_bound - lower_bound)
    data = np.clip(data, 0.0, 1.0, out=data)
    # data = exposure.rescale_intensity(data, in_range=(lower_bound, upper_bound))
    logging.info("Converting to uint8.")
    data = np.multiply(data, 255, out=data)
    return data.astype(np.uint8)


def get_num_of_ims(vol_shape: Tuple, axis_enum: Axis):
    """Calculates the number of images that will be created when slicing
    an image volume in specified planes or combinations of planes.

    Args:
        vol_shape (tuple): 3d volume shape (z, y, x).
        axis_enum (Enum): Defines specific axis or all axes

    Returns:
        int: Total number of images that will be created when the volume is
        sliced.
    """
    if axis_enum == Axis.ALL:
        return sum(vol_shape)
    else:
        return vol_shape[axis_enum.value]


def get_axis_index_pairs(vol_shape: Tuple, axis_enum: Axis):
    """Gets all combinations of axis and image slice indices that are found
    in a 3d volume.

    Args:
        vol_shape (tuple): 3d volume shape (z, y, x)
        axis_enum (Enum): Defines specific axis or all axes

    Returns:
        itertools.chain: An iterable containing all combinations of axis
        and image index that are found in the volume.
    """

    if axis_enum == Axis.ALL:
        return chain(
            product("z", range(vol_shape[0])),
            product("y", range(vol_shape[1])),
            product("x", range(vol_shape[2])),
        )
    else:
        return product(axis_enum.name.lower(), range(vol_shape[axis_enum.value]))


def axis_index_to_slice(vol, axis, index):
    """Converts an axis and image slice index for a 3d volume into a 2d
    data array (slice).

    Args:
        vol (3d array): The data volume to be sliced.
        axis (str): One of 'z', 'y' and 'x'.
        index (int): An image slice index found in that axis.

    Returns:
        2d array: A 2d image slice corresponding to the axis and index.
    """
    if axis == "z":
        return vol[index, :, :]
    if axis == "y":
        return vol[:, index, :]
    if axis == "x":
        return vol[:, :, index]


def save_data_to_hdf5(data, file_path, internal_path="/data", chunking=True):
    logging.info(f"Saving data of shape {data.shape} to {file_path}.")
    with h5.File(file_path, "w") as f:
        f.create_dataset(
            internal_path, data=data, chunks=chunking, compression=cfg.HDF5_COMPRESSION
        )



__all__ = ["get_2d_training_parser", "get_2d_prediction_parser", "Quality"]


from volume_segmantics.utilities.arg_parsing import (
    get_2d_prediction_parser,
    get_2d_training_parser,
)
from volume_segmantics.utilities.base_data_utils import Quality
